/*
 * Copyright (C) 1997-2005 Andre Noll <noll@mathematik.tu-darmstadt.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */


#include "server.h"

extern struct misc_meta_data *mmd;
extern struct dbtool dblist[];

static FILE *audio_file = NULL;
static char supported_audio_formats[MAXLINE];

#if 1
	int mp3_init(void *);
#endif

#ifdef HAVE_OGGVORBIS
	int ogg_init(void *);
#endif

static struct audio_format afl[] = { /* audio format list */
#if 1
	{
		.name = "mp3",
		.init = mp3_init,
		.sw_pid = 0,
		.fifo = -1
	},
#endif
#ifdef HAVE_OGGVORBIS
	{
		.name = "ogg",
		.init = ogg_init,
		.sw_pid = 0,
		.fifo = -1
	},
#endif
	{
		.name = NULL,
	}
};

const char *audio_format_name(int i)
{
	return i >= 0?  afl[i].name : "(none)";
}

char *afs_stream_writer_cmd(int i)
{
	return i >=0? afl[i].open_cmd : "/bin/false";
}


char *afs_supported_audio_formats(void)
{
	return supported_audio_formats;
}

void afs_handle_sigchild(pid_t pid)
{
	int i;
	for (i = 0; afl[i].name; i++) {
		if (afl[i].sw_pid == pid) {
			clog(INFO, "%s: %s stream writer (pid %d) died, \n", 
				__func__, afl[i].name, pid);
			afl[i].sw_pid = 0;	
		}
	}
}

/*
 * Set close on exec flag after opening fifo to mark its fd for closing
 * after fork. Otherwise closing the pipe does not have the desired 
 * effect as the child of the server still holds open the fifo.
 */
int afs_open_fifo(char *fifoname)
{
	int ret, fifo;
	struct stat stat_buf;

	ret = stat(fifoname, &stat_buf);
	if (ret < 0) {
		ret = mknod(fifoname, S_IFIFO | S_IRUSR | S_IWUSR, 42);
		if (ret < 0) {
			clog(CRIT, "%s: can not create fifo %s (%s)\n", __func__,
				fifoname, strerror(errno));
			return -1;
		}
		ret = stat(fifoname, &stat_buf);
	}
	if (!S_ISFIFO(stat_buf.st_mode)) {
		clog(CRIT, "%s: fifo file %s exists, but is not a fifo\n", 
			__func__, fifoname, strerror(errno));
		return -1;
	}
	fifo = open(fifoname, O_WRONLY);
	if (fifo < 0) {
		clog(CRIT, "%s: fifo open error (%s)\n", __func__, 
			strerror(errno));
		return -1;
	}
	clog(DEBUG, "%s: %s opened (%d)\n", __func__, fifoname, fifo);
	ret = fcntl(fifo, F_SETFD, FD_CLOEXEC);
	ret = fcntl(fifo, F_GETFD);
	return fifo;
}

static int setup_stream_command(struct gengetopt_args_info *conf, int num)
{
	int i;
	struct audio_format *af = &afl[num];
	const char *n = af->name;
	int g = conf->stream_write_cmd_given;

	af->open_cmd = NULL;
	if (!g)
		goto out;
	clog(DEBUG, "%s: %d stream write command%s given, checking for %s\n", 
		__func__, g, g > 1? "s" : "", n);
	for (i = g - 1; i >= 0; i--) {
		char *c = conf->stream_write_cmd_arg[i];
		size_t len = strlen(n);
		if (strlen(c) <= len || strncmp(c, n, len))
			continue;
		if (*(c + len) != ':') {
			clog(WARNING, "%s: invalid stream_write_cmd\n", __func__);
			return -1;
		}
		af->open_cmd = c + len + 1;
		clog(INFO, "%s: using \"%s\" for files of type %s\n", 
			__func__, af->open_cmd, n);
		return 1;
	}
out:
	return 0;
}

int afs_handle_sighup(struct gengetopt_args_info *conf)
{
	int i, ret;
	for (i = 0; afl[i].name; i++)
		if ((ret = setup_stream_command(conf, i)) < 0)
			return ret;
	return 1;
}

int afs_init(struct gengetopt_args_info *conf)
{
	int i, ret;
	char *hn = getenv("HOSTNAME"), *home = getenv("HOME");

	clog(INFO, "%s: checking audio formats\n", __func__);
	*supported_audio_formats = '\0';
	for (i = 0; afl[i].name; i++) {
		clog(DEBUG, "%s: supported audio format: %s\n", __func__, 
			afl[i].name);
		if ((ret = setup_stream_command(conf, i)) < 0)
			return ret;
		afl[i].fifoname = make_message("%s/.paraslash/server_%s_fifo.%s", 
			home? home : "/tmp", afl[i].name, 
			hn? hn : "unknown_host");
		if (afl[i].init(&afl[i]) < 0) {
			clog(ERROR, "%s: init %s failed, disabling "
				"%s support\n", __func__, afl[i].name, 
				afl[i].name);
			afl[i].get_file_info = NULL;
			continue;
		}
		sprintf(supported_audio_formats, "%s%s ", 
			*supported_audio_formats? supported_audio_formats: "",
			 afl[i].name);
	}
	if (!i) {
		clog(EMERG, "%s: no audio formats supported\n");
		return -1;
	}
	clog(DEBUG, "%s: supported audio formats: %s\n", __func__, 
		supported_audio_formats);
	
	return i;
}

static int get_file_info(int i)
{
	/* get_file_info must not close the audio file, even if it is invalid */
	int ret = afl[i].get_file_info(audio_file, mmd->audio_file_info,
		&mmd->chunks_total, &mmd->seconds_total);
	if (ret > 0)
		clog(INFO, "%s: valid %s file\n", __func__, afl[i].name);
	else
		clog(DEBUG, "%s: %s format not recognized\n", __func__, 
			afl[i].name);
	return ret;
}

static int guess_audio_format(const char *name)
{

	int i, len1 = strlen(name), len2;

	for (i = 0; afl[i].name; i++) {
		len2 = strlen(afl[i].name);
		if (len1 < len2 || !afl[i].get_file_info)
			continue;
		if (!strncasecmp(name + (len1 - len2), afl[i].name, len2)) {
			clog(DEBUG, "%s: might be %s\n", __func__,
				afl[i].name);
			return i;
		}
	}
	return -1;
}

static int get_audio_format(int omit)
{
	int i;

	for (i = 0; afl[i].name; i++) {
		if (i == omit || !afl[i].get_file_info)
			continue;
		rewind(audio_file);
		if (get_file_info(i) > 0)
			return i;
		rewind(audio_file);
	}
	return -1;
}

/*
 * upddate shared mem
 */
static int update_mmd(void)
{
	int i;
	struct stat file_status;

	i = guess_audio_format(mmd->filename);
	if (i < 0 || get_file_info(i) < 0)
		i = get_audio_format(i);
	if (i < 0) {
		clog(WARNING, "%s: unknown audio format for %s\n", __func__,
			mmd->filename);
		return -1;
	}
	mmd->audio_format = i;
	mmd->chunks_sent = 0;
	if (fstat(fileno(audio_file), &file_status) == -1) {
		clog(ERROR, "%s: fstat error (%s)\n", __func__, strerror(errno));
		return -1;
	}
	mmd->size = file_status.st_size;
	mmd->mtime = file_status.st_mtime;
	clog(NOTICE, "%s: next audio file: %s\n", __func__, mmd->filename);
	return 1;
}

static void get_song(void)
{
	char **sl = NULL;
	int i;

	if (mmd->afs_status_flags & AFS_NOMORE)
		goto err_out;

	sl = dblist[mmd->dbt_num].get_song_list(10);
	if (!sl)
		goto err_out;
	for (i = 0; sl[i]; i++) {
		clog(INFO, "%s: trying %s\n", __func__, sl[i]); 
		if (strlen(sl[i]) >= _POSIX_PATH_MAX)
			continue;
		audio_file = fopen(sl[i], "r");
		if (!audio_file)
			continue;
		strcpy(mmd->filename, sl[i]);
		if (update_mmd() < 0) {
			fclose(audio_file);
			continue;
		}
		mmd->num_played++;
		dblist[mmd->dbt_num].update_song(sl[i]);
		clog(DEBUG, "%s: success\n", __func__); 
		mmd->afs_status_flags &= (~AFS_NEXT);
		goto free;
	}
	clog(ERROR, "%s: no valid files found\n", __func__);
err_out:
	mmd->afs_status_flags = AFS_NEXT;
free:
	if (sl) {
		for (i = 0; sl[i]; i++)
			free(sl[i]);
		free(sl);
	}
}

int call_all_mainloops(struct timeval **next_time)
{
	struct timeval *dummy;
	int i, fd = -1, format = mmd->audio_format;
	unsigned int flags = mmd->afs_status_flags;

	*next_time = NULL;
	for (i = 0; afl[i].name; i++) {
//	clog(DEBUG, "%s: i = %d\n", __func__, i); 
		if (i == format)
			fd = afl[i].mainloop(1, flags, next_time);
		else
			afl[i].mainloop(0, flags, &dummy);
	}
	return fd;
}

/*  
 * < 0: time left. remainder was written to timeout
 * > 0: no next_time or next_time expired  
 */
static int afs_compute_timeout(struct timeval *next_time, struct timeval **timeout)
{
	static struct timeval the_timeout;
	struct timeval now;
	int to, ret = 1;

	gettimeofday(&now, NULL);
	if (!next_time || tv_diff(&now, next_time, NULL) >= 0)
		goto out;
	tv_diff(next_time, &now, &the_timeout);
	to = tv2ms(&the_timeout);
//	clog(DEBUG, "%s: %lims left\n", __func__, to);
	ret = -1;
	if (to > 10) {
		*timeout = &the_timeout;
		goto out;
	}
	ret = 1;
//	if (to > 0)
//		usleep(to * 1000);
out:
	return ret;
}

/*
 * called from main in any case, just before main calls select
 */
int afs_mainloop(struct timeval **timeout)
{
	struct audio_format *af = NULL;
	struct timeval *next_time;
	int ret = -1, sw_fileno, playing, next, format;
	
	*timeout = NULL;
again:
	playing = mmd->afs_status_flags & AFS_PLAYING;
	next = mmd->afs_status_flags & AFS_NEXT;
	format = mmd->audio_format;
	if (format >= 0)
		af = afl + format;
	ret = -1;
	sw_fileno = call_all_mainloops(&next_time);
	if (next) {
		audio_file = NULL;
		/* do nothing until stream writer dies */
		if (af && af->sw_pid)
			goto out;
	}
get_audio_file:
	if (!audio_file && !next_time && playing) {
		clog(DEBUG, "%s: ready and playing, but no audio file\n", 
			__func__);
		get_song();
		goto again;
	}
	if (!playing || !audio_file || !af || !af->sw_pid) {
		clog(DEBUG, "%s: not playing\n", __func__);
		goto out;
	}
	ret = afs_compute_timeout(next_time, timeout);
	if (ret > 0 && !audio_file) {
		next_time = NULL;
		goto get_audio_file;
	}
	if (ret < 0)
		goto out;
	ret = sw_fileno;
out:
//	clog(DEBUG, "%s: returning fd %d, tv = %lims\n", __func__,
//		ret, *timeout? tv2ms(*timeout) : 0); 
	return ret;
}

/*
 * Status is playing and stream writer is ready for writing.
 */
int afs_send_chunk(void)
{
	int ret;
	struct audio_format *af;

	if (mmd->audio_format < 0 || !audio_file) {
		clog(CRIT, "%s: BUG\n", __func__);
		return -1;
	}
	af = &afl[mmd->audio_format];
	ret = af->send_chunk(mmd->audio_file_info, 
		&mmd->chunks_sent, mmd->afs_status_flags);
	mmd->afs_status_flags &= ~(AFS_NEXT | AFS_REPOS);
	if (ret > 0)
		mmd->afs_status_flags |= AFS_PLAYING;
	if (ret < 0)
		mmd->afs_status_flags = AFS_NEXT;
	if (!ret)
		mmd->afs_status_flags |= AFS_NEXT;
	return ret;
}
